import logging
import asyncio
from datetime import datetime
import signal
import sys
from monstr.exception import ConfigurationError
from monstr.client.client import Client, ClientPool
from monstr.event.event import Event
from monstr.util import util_funcs
from monstr.encrypt import Keys
from monstr.signing import BasicKeySigner
from bots.basic import BotEventHandler


# default relay
DEFAULT_RELAY = 'ws://localhost:8081'
# default key - if None it'll be generated each run
USE_KEY = 'nsec1fnyygyh57chwf7zhw3mwmrltc2hatfwn0hldtl4z5axv4netkjlsy0u220'


class IPBot(BotEventHandler):

    async def make_response(self, client: Client, sub_id, evt: Event) -> Event:
        ret = await self.get_reply_event(evt)

        # any shell method to get ip will do us, we'll use curl - hopefully it exists
        proc = await asyncio.create_subprocess_exec(
            'dig', '@resolver4.opendns.com', 'myip.opendns.com', '+short',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE
        )

        stdout, stderr = await proc.communicate()

        ret.content = stdout.decode()
        return ret


def get_args() -> dict:
    ret = {
        'keys': USE_KEY,
        'relays': DEFAULT_RELAY
    }

    use_keys = Keys.get_key(ret['keys'])
    if use_keys is None or use_keys.private_key_hex() is None:
        raise ConfigurationError(f'{ret["keys"]} bad key value or public key only')

    ret['keys'] = use_keys

    return ret


async def run_bot(args):
    # just the keys, change to profile?
    keys: Keys = args['keys']

    # relays we'll watch
    relays = args['relays']

    # kinds will reply to
    kind = Event.KIND_ENCRYPT

    # actually create the client pool
    def on_connect(the_client: Client):
        the_client.subscribe(sub_id='bot_watch',
                             handlers=[bot],
                             filters={
                                 'kinds': [kind],
                                 '#p': [keys.public_key_hex()],
                                 'since': util_funcs.date_as_ticks(datetime.now())
                             })

    clients = ClientPool(clients=relays.split(','),
                         on_connect=on_connect)

    # actually create the bot
    bot = IPBot(signer=BasicKeySigner(keys),
                clients=clients,
                encrypt_kinds=[])

    # start the clients
    print(f'monitoring for events from or to account {keys.public_key_hex()} on relays {relays}')

    def sigint_handler(signal, frame):
        clients.end()
        sys.exit(0)

    signal.signal(signal.SIGINT, sigint_handler)

    await clients.run()


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.DEBUG)
    try:
        asyncio.run(run_bot(get_args()))
    except ConfigurationError as ce:
        print(ce)



import logging
import toml
import sys
from pathlib import Path
from toml import TomlDecodeError
import os


def load_toml(filename, dir):
    if os.path.sep not in filename:
        filename = dir+os.path.sep+filename

    ret = {}
    f = Path(filename)
    if f.is_file():
        try:
            ret = toml.load(filename)
        except TomlDecodeError as te:
            print(f'Error in config file {filename} -{te}')
            sys.exit(2)

    else:
        logging.debug(f'load_toml:: no config file {filename}')
    return ret



import logging
import json
from abc import ABC
import aiohttp
from aiohttp import ClientSession
from monstr.client.client import ClientPool
from monstr.signing import SignerInterface
from monstr.inbox import Inbox
from monstr.client.event_handlers import EventAccepter
from bots.basic import BotEventHandler, CommandMapper


class BitcoindRPC:

    def __init__(self,
                 url: str,
                 user: str,
                 password: str):

        self._url = url
        self._user = user
        self._password = password

    async def _execute_cmd(self, method, params):
        try:
            async with ClientSession() as session:
                print(self._url)
                async with session.post(
                        url=self._url,
                        data=json.dumps({
                            'method': method,
                            'params': params
                            # 'jsonrpc': '2.0',
                            # 'id': self._id
                        }),
                        auth=aiohttp.BasicAuth(self._user, self._password)
                ) as resp:
                    if resp.status == 200:
                        ret = json.loads(await resp.text())
                    else:
                        raise Exception(f'Bitcoind_rpc:: execute_cmd failed method- {method} params - {params} status {resp.status}')
                        # logging.debug(f'Bitcoind_rpc:: execute_cmd failed method- {method} params - {params} status {resp.status}')

        except Exception as e:
            ret = {
                'error': str(e)
            }

        return ret

    async def getbalances(self):
        return await self._execute_cmd('getbalances', {})

    async def getnewaddress(self):
        return await self._execute_cmd('getnewaddress',{})

    async def listtransactions(self):
        return await self._execute_cmd('listtransactions',{})

    async def listunspent(self):
        return await self._execute_cmd('listunspent',{})

    async def sendrawtransaction(self, hexstring: str):
        return await self._execute_cmd('sendrawtransaction', [hexstring])


class BitcoindCommandMapper(CommandMapper, ABC):

    def __init__(self, bitcoind: BitcoindRPC):
        self._rpc = bitcoind
        super().__init__({
            'getbalances': self.getbalances,
            'getnewaddress': self.getnewaddress,
            'listtransactions': self.listtransactions,
            'listunspent': self.listunspent,
            'sendrawtransaction' : self.sendrawtransaction
        })

    def is_cmd_auth(self, name, pub_k: str) -> bool:
        return True

    async def getbalances(self, args):
        return await self._rpc.getbalances()

    async def getnewaddress(self, args):
        return await self._rpc.getnewaddress()

    async def listtransactions(self, args):
        return await self._rpc.listtransactions()

    async def listunspent(self, args):
        return await self._rpc.listunspent()

    async def sendrawtransaction(self, args):
        return await self._rpc.sendrawtransaction(args[0])


class BitcoindBot(BotEventHandler):

    def __init__(self,
                 signer: SignerInterface,
                 clients: ClientPool,
                 bitcoin_rpc: BitcoindRPC,
                 kind: int = 20888,
                 inbox: Inbox = None,
                 event_acceptors: [EventAccepter] = None):

        self._rpc = bitcoin_rpc

        self._kind = kind

        super().__init__(signer=signer,
                         clients=clients,
                         kinds=[kind],
                         encrypt_kinds=[kind],
                         inbox=inbox,
                         event_acceptors=event_acceptors,
                         command_map=BitcoindCommandMapper(bitcoin_rpc))

    @property
    def kind(self):
        return self._kind




import logging
import asyncio
from datetime import datetime
import signal
import sys
from pathlib import Path
from monstr.exception import ConfigurationError
from monstr.client.client import Client, ClientPool
from monstr.util import util_funcs
from monstr.encrypt import Keys
from monstr.signing import BasicKeySigner
from bots.bitcoind import BitcoindBot, BitcoindRPC
from src.bots.util import load_toml
from bots.acceptors import AuthListAccept


# working directory
WORK_DIR = f'{Path.home()}/.nostrpy/'
# config file
CONFIG_FILE = f'bitcoin_bot.toml'
# default relay
DEFAULT_RELAY = 'ws://localhost:8081'
# default key - if None it'll be generated each run
USE_KEY = 'nsec1fnyygyh57chwf7zhw3mwmrltc2hatfwn0hldtl4z5axv4netkjlsy0u220'

# for connecting to bitcoind
BITCOIND_NETWORK = 'test'            #   not yet used have to manually change port
# probably better this comes from the toml file....
BITCOIND_USER = None
BITCOIND_PASSWORD = None

# default bitcoind config
BITCOIND_WALLET = 'cormorant'
BITCOIND_HOST = 'http://localhost'
BITCOIND_PORT = 8332
# BITCOIND_PORT = 18332


def get_args() -> dict:
    ret = {
        'work-dir': WORK_DIR,
        'conf': CONFIG_FILE,
        'keys': USE_KEY,
        'relays': DEFAULT_RELAY,
        'bitcoind-host': BITCOIND_HOST,
        'bitcoind-wallet': BITCOIND_WALLET,
        'bitcoind-port': BITCOIND_PORT,
        'bitcoind-user': BITCOIND_USER,
        'bitcoind-password': BITCOIND_PASSWORD
    }

    # update from toml file
    ret.update(load_toml(ret['conf'], ret['work-dir']))

    # TODO - parse args cli
    # ret.update(get_cmdline_args(ret))

    use_keys = Keys.get_key(ret['keys'])
    if use_keys is None or use_keys.private_key_hex() is None:
        raise ConfigurationError(f'{ret["keys"]} bad key value or public key only')

    ret['keys'] = use_keys

    return ret


async def run_bot(args):
    # just the keys, change to profile?
    keys: Keys = args['keys']

    # relays we'll watch
    relays = args['relays']

    # bitcoin connection stuff - note this will probably become bitcoin_ after we add argparse
    bitcoin_host = args['bitcoind-host']
    bitcoin_port = args['bitcoind-port']
    # this should probably just be default wallet to use (or always expected passed in on call?)
    bitcoin_wallet = args['bitcoind-wallet']
    bitcoin_url = f'{bitcoin_host}:{bitcoin_port}/wallet/{bitcoin_wallet}'
    bitcoin_user = args['bitcoind-user']
    bitcoin_password = args['bitcoind-password']

    # actually create the client pool
    def on_connect(the_client: Client):
        watch_filter = {
             'kinds': [bot.kind],
             '#p': [keys.public_key_hex()],
             'since': util_funcs.date_as_ticks(datetime.now())
        }
        if bot.inbox:
            watch_filter = {
                'kinds': [bot.inbox.kind],
                'authors': [bot.inbox.pub_key],
                'since': util_funcs.date_as_ticks(datetime.now())
            }
        print(watch_filter)
        the_client.subscribe(sub_id='bot_watch',
                             handlers=[bot],
                             filters=watch_filter)

    clients = ClientPool(clients=relays.split(','),
                         on_connect=on_connect)

    # accept request from these keys - this could come from file/config
    # or even from monitoring a nostr list event
    my_accept = AuthListAccept([
        '5c4bf3e548683d61fb72be5f48c2dff0cf51901b9dd98ee8db178efe522e325f'
    ])


    inbox = None
    # need to get cmd
    # inbox = Inbox(keys=Keys.get_key('nsec1g6exndqqcrmwxrkhkn0d4xer9hvr3nmnd8umm4kf2ml7vrqvgl7qyccx5k'),
    #               use_kind=20888)
    # inbox.set_share_map(for_keys=keys,
    #                     to_keys=[Keys.get_key('npub1t39l8e2gdq7kr7mjhe053skl7r84ryqmnhvca6xmz780u53wxf0swj0fey')])


    # actually create the bot
    bot = BitcoindBot(signer=BasicKeySigner(keys),
                      clients=clients,
                      bitcoin_rpc=BitcoindRPC(
                          url=bitcoin_url,
                          user=bitcoin_user,
                          password=bitcoin_password
                      ),
                      inbox=inbox,
                      event_acceptors=[my_accept])

    # start the clients
    print(f'monitoring for events from or to account {keys.public_key_hex()} on relays {relays}')
    def sigint_handler(signal, frame):
        clients.end()
        sys.exit(0)

    signal.signal(signal.SIGINT, sigint_handler)
    await clients.run()


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.DEBUG)
    try:
        asyncio.run(run_bot(get_args()))
    except ConfigurationError as ce:
        print(ce)



import logging
import asyncio
import json
from datetime import datetime
from abc import ABC, abstractmethod
from monstr.client.client import Client, ClientPool
from monstr.client.event_handlers import EventHandler, DeduplicateAcceptor, EventAccepter
from monstr.signing import SignerInterface
from monstr.event.event import Event
from monstr.inbox import Inbox
from monstr.util import util_funcs


class CommandMapper(ABC):

    def __init__(self, command_map: dict):
        self._command_map = command_map

    def is_command(self, name: str) -> bool:
        return name in self._command_map

    @abstractmethod
    def is_cmd_auth(self, name, pub_k: str) -> bool:
        # TODO - this or similar allows us to restrict certain commands
        pass

    async def do_command(self, name: str, args: []) -> dict:
        return await self._command_map[name](args)


class BotEventHandler(EventHandler):

    def __init__(self,
                 signer: SignerInterface,
                 clients: ClientPool,
                 # default is to reply both to standard notes and nip 4 dms
                 kinds: [int] = None,
                 # the encrypted kinds
                 encrypt_kinds: [int] = None,
                 inbox: Inbox = None,
                 event_acceptors: [EventAccepter] = None,
                 command_map: CommandMapper = None):

        # used to sign and encrypt events
        self._signer = signer

        # the relays we're watching for events from and writing to
        # (can be a pool with different relays for reading to writing)
        self._clients = clients

        # if not given default is to reply both to plaintext and nip4 encrypted events
        if kinds is None:
            kinds = [Event.KIND_TEXT_NOTE, Event.KIND_ENCRYPT]
        # the event kind/s we'll use to recieve message - and probably to write also default kind 1 text notes
        elif isinstance(kinds, list):
            kinds = set(kinds)
        self._kinds = kinds

        # events that'll be treated as nip4 encryped, default is just Event.KIND_ENCRYPT(4)
        if encrypt_kinds is None:
            encrypt_kinds = [Event.KIND_ENCRYPT]
        elif isinstance(encrypt_kinds, list):
            encrypt_kinds = set(encrypt_kinds)
        self._encrypt_kinds = encrypt_kinds

        # public inbox to use - not implemented yet
        self._inbox = inbox

        # acceptance code run before we'll accept and event, by default just de-duplication
        self._acceptors = [DeduplicateAcceptor()]
        if event_acceptors:
            self._acceptors = self._acceptors + event_acceptors

        # map commands to functions - if not given
        self._command_map = command_map

        super().__init__(event_acceptors=self._acceptors)

    @staticmethod
    def reply_tags(src_evt: Event) -> []:
        """
            returns the minimal reply tags that bot should make on an event its responding too
            TODO - make sure we deal with root reply correctly based on src_event
        """
        return [
            ['p', src_evt.pub_key],
            ['e', src_evt.id, 'reply']
        ]

    def do_event(self, client: Client, sub_id, evt: Event):
        # move into async, probably put in q here
        asyncio.create_task(self.ado_response_event(client=client,
                                                    sub_id=sub_id,
                                                    evt=evt))

    async def ado_response_event(self, client: Client, sub_id, evt: Event) -> Event:
        # if inbox unwrap
        if self._inbox:
            evt = await self._inbox.unwrap_event(evt=evt,
                                                 user_sign=self._signer)

        # replying to ourself would be bad! also call accept_event
        # to stop us replying mutiple times if we see the same event from different relays
        if evt.pub_key == await self._signer.get_public_key() or \
                self.accept_event(client, sub_id, evt) is False \
                or evt.kind not in self._kinds:
            return None

        # decode/unwrap or whatever to get the actual event with the cmd in it
        evt = await self.get_request_event(evt)

        # NOTE at this point is the decrypted and unwrapped event (if that was required)
        if self._command_map is None:
            response_evt = await self.make_response(client, sub_id, evt)
        else:
            response_evt = await self.make_response_cmd_map(client, sub_id, evt)

        # we have the response - this just encrypts/wraps for sending
        response_evt = await self.make_response_event(response_evt, evt)

        # and actually send
        await self.send_response(response_evt)

        return response_evt

    # def authorise(self, evt: Event) -> bool:
    #     # auth check on basic event, do we even bother responding to this user?
    #     return True

    async def get_request_event(self, evt: Event) -> Event:
        ret = evt
        # TODO - add inbox unwrapping if required here

        # decrypt if required
        if evt.kind in self._encrypt_kinds:
            ret = await self._signer.nip4_decrypt_event(evt)

        return ret

    async def make_response_event(self, response_evt: Event, prompt_evt: Event) -> Event:
        # should be the adding whatever we strip off during get cmd event
        ret = response_evt
        # TODO - add inbox wrapping if required here

        # encrypt if required
        if response_evt.kind in self._encrypt_kinds:
            ret = await self._signer.nip4_encrypt_event(evt=response_evt,
                                                        to_pub_k=prompt_evt.pub_key)


        return ret

    async def get_reply_event(self, prompt_evt: Event):
        # gets an empty reply event ready for us to set content - may also add tags
        return Event(
            # reply will be same kind that we recieved on
            kind=prompt_evt.kind,
            content='',
            tags=BotEventHandler.reply_tags(prompt_evt),
            pub_key=await self._signer.get_public_key(),
            created_at=util_funcs.date_as_ticks(datetime.now())
        )

    async def make_response(self, client: Client, sub_id, evt: Event) -> Event:
        ret = await self.get_reply_event(evt)
        ret.content = 'BotEventHandler:: No CommandMapper? You should implement make_response in your bot'
        return ret

    async def make_response_cmd_map(self, the_client: Client, sub_id, evt: Event) -> Event:
        cmd_text = evt.content
        cmd_split = cmd_text.split()
        response_text = 'no command given?!'

        if cmd_split:
            cmd = cmd_split[0]
            args = cmd_split[1:]
            if not self._command_map.is_command(cmd):
                response_text = json.dumps({
                    'error': f'command not understood - {cmd}'
                })
            elif not self._command_map.is_cmd_auth(cmd, evt.pub_key):
                response_text = json.dumps({
                    'error': f'command not understood - {cmd}'
                })
            else:
                response_text = json.dumps(await self._command_map.do_command(cmd, args))

        ret = await self.get_reply_event(evt)
        ret.content = response_text
        return ret

    async def send_response(self, response_evt: Event):
        await self._signer.sign_event(response_evt)

        # wrap the response event
        if self._inbox:
            response_evt = await self._inbox.wrap_event(evt=response_evt,
                                                        from_sign=self._signer)

        self._clients.publish(response_evt)

    @property
    def inbox(self) -> Inbox:
        return self._inbox



from monstr.client.client import Client
from monstr.client.event_handlers import EventAccepter
from monstr.event.event import Event
from monstr.encrypt import Keys


class AuthListAccept(EventAccepter):
    """
        basic auth acceptor that will only return True if evt pub_k is in accept_pub_k
        this could come from file or better a nostr list event (we could then watch for updates of the list)
    """
    def __init__(self, accept_pub_k: [str | Keys] = None):
        self._accept = accept_pub_k
        if self._accept is not None:
            self._accept = self._get_keys(self._accept)

    def _get_keys(self, key_list: []) -> set:
        ret = set()
        for c_i in key_list:
            k = c_i
            if not isinstance(k, Keys):
                k = Keys.get_key(k)
            if k is not None:
                ret.add(k.public_key_hex())

        print(ret)
        return ret



    def accept_event(self, the_client: Client, sub_id: str, evt: Event) -> bool:
        print(self._accept, evt.pub_key)
        ret = False
        if self._accept is None or evt.pub_key in self._accept:
            ret = True
        return ret
